import java.util.HashMap;
import java.util.Map;
import java.util.Vector;

public class Main
{
//    public static Rational eps = Rational.one;
    public static Polynomial eps = new Polynomial(new Rational(1,100));//new Polynomial(new Rational(1,100));//new Polynomial("__eps");
    public static Polynomial C = new Polynomial("__C");
    public static int mu = 2, degree = 0;

    public static String fileName1 = "", fileName2="", solver = "", workingdir="",
            solversDir="", varType="", distance_or_equivalence= "",OST_condition = "", sting = "",aspic="";

    public static void main(String[] args) throws Exception
    {
        long startTime = System.currentTimeMillis();

        fileName1 = args[0];
        fileName2 = args[1];
        solver = "gurobi";
        workingdir = "work";
        solversDir = "solvers";
        varType = "Real"; // Real or Int
        degree = Integer.parseInt(args[2]);
        mu = degree;
        distance_or_equivalence = args[3];
        OST_condition = args[4];
        sting=args[5];
        aspic = args[6];
        StringBuilder params = new StringBuilder();
        for(int i=2;i<=6;i++)
            params.append(args[i]).append(" ");
        System.out.println(params);

        System.err.println("------------Reading First Transition System T1------------");
        TransitionSystem T1 = Parser.parse(fileName1,"input1");
        fileName1 = fileName1.replace("/", "_");

        System.err.println("------------Reading Second Transition System T2------------");
        TransitionSystem T2 = Parser.parse(fileName2,"input2");
        fileName2 = fileName2.replace("/", "_");

        System.err.println("------------Verifying T1 transitions------------");
        T1.verify();
        T1.config_observe();
        System.err.println("------------Verifying T2 transitions------------");
        T2.verify();
        T2.config_observe();
        if(!aspic.equals("noaspic"))
        {
            System.err.println("------------Extracting Invariants for T1 from ASPIC------------");
            T1.get_aspic_invariants();
            for (CFGNode n : T1.states)
                if (T1.aspic_invariant != null && T1.aspic_invariant.get(n.id) != null)
                    System.err.println("State: " + n.id + "  ASPIC-Invariant: " + T1.aspic_invariant.get(n.id).toNormalString());
            System.err.println("------------Extracting Invariants for T2 from ASPIC------------");
            T2.get_aspic_invariants();
            for (CFGNode n : T2.states)
                if (T2.aspic_invariant != null && T2.aspic_invariant.get(n.id) != null)
                    System.err.println("State: " + n.id + "  ASPIC-Invariant: " + T2.aspic_invariant.get(n.id).toNormalString());
        }
        if(!sting.equals("nosting"))
        {
            System.err.println("------------Extracting Invariants for T1 from STING------------");
            T1.get_sting_invariants();
//        for(CFGNode n: T2.states)
//            System.err.println("State: "+n.id+"  Sting-Invariant: "+T1.sting_invariant.get(n.id).toNormalString());

            System.err.println("------------Extracting Invariants for T2 from STING------------");
            T2.get_sting_invariants();
        }
//
        System.err.println("------------Setting Template for Cost Function------------");
        Polynomial cost = Martingale.MakeCostTemplate();
        System.err.println("Cost: "+cost.toNormalString());
        System.err.println("------------Forcing Cost Function To Be 1-Lipschitz on T1------------");
        Handelman lipschitz1 = Martingale.lipschitz_condition(T1,cost);
        System.err.println("------------Forcing Cost Function To Be 1-Lipschitz on T2------------");
        Handelman lipschitz2 = Martingale.lipschitz_condition(T2,cost);
        System.err.println("------------Setting Template for UESM of T1------------");
        Martingale.MakeTemplate(T1);
        for(CFGNode n:T1.states)
            System.err.println("State: "+n.id+"  -->  "+n.rank.toNormalString());
        System.err.println("------------Setting Template for LESM of T2------------");
        Martingale.MakeTemplate(T2);
        for(CFGNode n:T2.states)
            System.err.println("State: "+n.id+"  -->  "+n.rank.toNormalString());



        Vector<Handelman> handelmanVector = Martingale.generate(T1,T2,cost,OST_condition);
        if(distance_or_equivalence.equals("distance"))
        {
            handelmanVector.add(lipschitz1);
            handelmanVector.add(lipschitz2);
        }
        Vector<Polynomial> equalities = new Vector<>();
        for(Handelman h:handelmanVector)
        {
            System.err.println(h.toString());
            equalities.addAll(h.generate_equalities());
        }
        System.err.println("------------Running "+solver+" on generated system of inequalities------------");
        double result = Handelman.run_solver(T1,T2,equalities,solver,handelmanVector,cost);
        if(distance_or_equivalence.equals("equivalence"))
        {
            if (result > 0.001)
                System.out.println("Successfully refuted equivalence!");
            else
                System.out.println("Failed");
        }
        else
        {
            if(solver.equals("gurobi") && result>0.000001)
            {
                System.out.println("Succesfully refuted similarity");
                System.out.println("Distance found: " + result);
            }
            else if(result>0.1)
                System.out.println("Succesfully refuted similarity");
            else
                System.out.println("Failed "+result);
        }
        long endTime = System.currentTimeMillis();
        System.out.println("total time used: " + (endTime - startTime));
    }
}